#!/usr/local/bin/php -q
<?

/*

Nanoweb, the aEGiS PHP web server
=================================

Copyright (C) 2002 Vincent Negrier aka. sIX <six@aegis-corp.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

error_reporting(E_PARSE | E_ERROR);

define(VERSION, "1.9.0");

define(DEFAULT_CONFIG_FILE, "/etc/nanoweb/nanoweb.conf");
define(DEFAULT_LISTEN_ADDR, "0.0.0.0");
define(DEFAULT_LISTEN_PORT, 80);
define(DEFAULT_LISTEN_QUEUE, 20);
define(DEFAULT_MIMETYPES, "/etc/mime.types");
define(DEFAULT_CONTENT_TYPE, "text/plain");
define(DEFAULT_REQUEST_TIMEOUT, 15);
define(DEFAULT_ACCESS_FILE, ".nwaccess");
define(DEFAULT_DOCROOT, "./");
define(DEFAULT_LOGFILE, "./access.log");
define(DEFAULT_FBWELCOMEFILE, ".welcome");
define(DEFAULT_STATIC_BUFFER_SIZE, 1048576);
define(DEFAULT_CONTENT_HANDLER, "static");

define(SCK_WRITE_PACKET_SIZE, 4096);
define(SCK_READ_PACKET_SIZE, 4096);
define(SCK_READ_SELECT_TIMEOUT, 2);

define(HTTP_VERSION, "HTTP/1.1");

define(SERVER_STRING, "aEGiS_nanoweb/".VERSION);
define(GENPAGE_SIGNATURE, "Page generated by <b>".SERVER_STRING."</b> at http://%SERVERNAME");

define(INT_MSGSIZE, 4096);

define(NM_HIT, "  HIT");
define(NM_RESTART_LOGGERS, "LGRST");
define(NM_SERVER_STATE, "SRVST");

define(NW_PLF_MAGIC, "DTC");
define(NM_SEP_MAGIC, md5(NW_PLF_MAGIC));
define(NM_SEP_LEN, strlen(NM_SEP_MAGIC));

define(NW_BAD_OUTSIDE_DOCROOT, 1);
define(NW_BAD_DOT_FILE, 2);

define(REQUIRED_PHP_VERSION, "4.2.0");

$HTTP_HEADERS=array(	200 => "200 OK",
			204 => "204 No Response",
			206 => "206 Partial Content",
			300 => "300 Multiple Choices",
			301 => "301 Moved Permanently",
			302 => "302 Found",
			303 => "303 See Other",
			304 => "304 Not Modified",
			307 => "307 Temporary Redirect",
			400 => "400 Bad Request",
			401 => "401 Unauthorized",
			403 => "403 Forbidden",
			404 => "404 Not Found",
			405 => "405 Method Not Allowed",
			406 => "406 Not Acceptable",
			408 => "408 Request Timeout",
			410 => "410 Gone",
			415 => "415 Unsupported Media Type",
			416 => "416 Requested Range Not Satisfiable",
			500 => "500 Internal Server Error",
			501 => "501 Not Implemented",
			503 => "503 Service Unavailable",
			506 => "506 Variant Also Negotiates");

$TEST_FUNCS=array(	"pcntl_fork"	=> false, 
			"socket_create"	=> true, 
			"posix_setuid"	=> false, 
			"gzencode"	=> false);

$conf_defaults=array(	"listeninterface"	=> DEFAULT_LISTEN_ADDR,
			"listenport"		=> DEFAULT_LISTEN_PORT,
			"listenqueue"		=> DEFAULT_LISTEN_QUEUE,
			"mimetypes"		=> DEFAULT_MIMETYPES,
			"requesttimeout"	=> DEFAULT_REQUEST_TIMEOUT,
			"accessfile"	=> DEFAULT_ACCESS_FILE,
			"documentroot"		=> DEFAULT_DOCROOT,
			"log"			=> DEFAULT_LOGFILE,
			"fbwelcomefile" 	=> DEFAULT_FBWELCOMEFILE,
			"defaultcontenttype" => DEFAULT_CONTENT_TYPE,
			"staticbuffersize" => DEFAULT_STATIC_BUFFER_SIZE,
			"defaulthandler" => DEFAULT_CONTENT_HANDLER);

$conf_vhosts_propagate=array(	"documentroot",
			"directoryindex",
			"serveradmin",
			"user",
			"group",
			"log",
			"filebrowser",
			"fbshowdotfiles",
			"fbwelcomefile",
			"userdir",
			"ignoredotfiles");

// Static response class

class static_response {

	var $content_length;

	function static_response($str) {

		$this->str=$str;
		$this->content_length=strlen($str);
	
	}

	function parser_open($args, $filename, &$rq_err, &$cgi_headers, &$add_ctype) {

	}

	function parser_get_output() {

		$s=$this->str;
		$this->str="";
		return($s);

	}

	function parser_eof() {

		return(!$this->str);
	
	}

	function parser_close() {

	}

}

$null_response =& new static_response("");
$lf=$null_response;

// Functions

function _parseconfig_close_part(&$conf, $cpart) {

	global $conf_defaults, $conf_vhosts_propagate;
	
	if ($cpart=="global") {
		
		foreach ($conf_defaults as $key=>$dval) if (!isset($conf[$cpart][$key])) $conf[$cpart][$key][0]=$dval;

	} else {

		foreach ($conf_vhosts_propagate as $key=>$pkey) if (!$conf[$cpart][$pkey]) $conf[$cpart][$pkey]=$conf["global"][$pkey];
		if (!$conf[$cpart]["servername"]) $conf[$cpart]["servername"][]=$cpart;

	}
	
	// Transform some directives (Dir = Idx Value)
	
	foreach (array("parseext", "errordocument", "errorheader") as $dir) {
	
		if ($conf[$cpart][$dir]) {

			foreach ($conf[$cpart][$dir] as $ps) {

				$ext=strtolower(substr($ps, 0, strpos($ps, " ")));
				$dext=trim(substr($ps, strpos($ps, " ")));

				$conf[$cpart]["_".$dir][$ext]=$dext;

			}
		
		}

	}
	
	return(true);

}

function _parseconfig_parse_line($str) {

	$cnfl=array();
	ereg("([^ =\n]+) *=? *([^\n]+)", $str, $cnfl);
	return(array(strtolower(trim($cnfl[1])), trim($cnfl[2])));

}

function parseconfig($conf_arr, $nodefaults=false) {

	$included_confs[$GLOBALS["conffile"]]=true;
	
	if ($clen=count($conf_arr)) {
		
		// Pass 1

		$key=-1;
		
		while ($key++<=$clen) {

			$str=$conf_arr[$key];
			
			list($cnfk, $cnfv)=_parseconfig_parse_line($str);

			if ($cnfk=="include") {

				if ($included_confs[$cnfv]) {

					errexit("configuration includes loop detected line ".$key." : '".trim($str)."'");
			
				}
			
				if (@!is_readable($cnfv)) {

					errexit("unable to include configuration file line ".$key." : '".$cnfv."'");
				
				}

				$subconf_arr=file($cnfv);

				$conf_arr=array_merge(array_slice($conf_arr, 0, $key), $subconf_arr, array_slice($conf_arr, $key+1));
				$key=-1;
				$clen=count($conf_arr);
				$included_confs[$cnfv]=true;
			
			}

		}

		// Pass 2
		
		foreach ($conf_arr as $key=>$str) {

			switch ($str[0]) {

				case "#":
				case ";":
				case "\n":
				break;
				
				case "[":
				if ($cpart && !$nodefaults) _parseconfig_close_part($conf, $cpart); 
				$cpart=substr(trim($str), 1, -1);
				unset($conf[$cpart]);
				break;

				default:

				list($cnfk, $cnfv)=_parseconfig_parse_line($str);

				if (trim($cnfk)) {
					
					if (!$cpart) $cpart="global";
					
					switch ($cnfk) {

						case "documentroot": 
						$conf[$cpart][$cnfk][]=$rp=realpath($cnfv);
						if (!@is_dir($rp)) errexit("directory not found at line ".$key." : '".trim($str)."'");
						break;

						case "serveralias": 
						if ($cnfv!=$cpart) $conf[$cnfv]=&$conf[$cpart];
						break;
						
						case "alias":
						$conf[$cpart][$cnfk][]=$cnfv;
						$aliases=explode(" ", $cnfv);
						$conf[$cpart]["_aliases"][$aliases[0]]=$aliases[1];
						break;
						
						default: 
						$conf[$cpart][$cnfk][]=$cnfv;
						break;

					}
				
				}

			}
			
		}

	}
	
	if (!$nodefaults) _parseconfig_close_part($conf, $cpart);
	
	return($conf);

}

function cmdline_conf_upd($conf, $cmdline_conf_overrides, $cmdline_conf_adds) {

	foreach ($cmdline_conf_overrides as $cs) {

		$ca=explode("=", $cs);
		$conf["global"][strtolower($ca[0])]=array($ca[1]);

	}

	foreach ($cmdline_conf_adds as $cs) {

		$ca=explode("=", $cs);
		$conf["global"][strtolower($ca[0])][]=$ca[1];

	}

	return($conf);

}

function load_modules($conf) {

	global $stats_modlist;
	
	if ($lm_arr=$conf["global"]["loadmodule"]) while (list($key, $modname)=each($lm_arr)) {

		$clsname=basename($modname, ".php");
		
		if (!is_file($modname)) {

			foreach (access_query("modulesdir") as $md) if (is_file($md."/".$modname)) {

				$moddir=$md;
				break;
			
			}
			
		} else $moddir="";
		
		if (!$ld_clss[$clsname]) {
			
			$nload=(!class_exists($clsname));
			
			@include_once($moddir.$modname);
			$modloaded=class_exists($clsname);

			if ($modloaded) {
			
				$ld_clss[$clsname]=true;
				$tmp=&new $clsname;
				$modules[$tmp->modtype][]=&$tmp;
				$tmp_modlist[]=array($clsname, $tmp->modname);
				if (is_array($tmp->urls)) while (list($k, $url)=each($tmp->urls)) $modules["url:".$url]=&$tmp;
				if ($nload) techo("loaded module : ".$tmp->modname);

			} else {

				techo("WARN: unable to load module '".$modname."'");
			
			}
	
		}
	
	}

	$stats_modlist=$tmp_modlist;
	return($modules);

}

function modules_init() {

	global $modules;
	
	foreach ($modules as $modclass)	for ($a=0;$a<count($modclass);$a++) {
		
		if ((method_exists($modclass[$a], "init")) && (!$modinit[$mc=get_class($modclass[$a])])) {
			
			$modclass[$a]->init();
			$modinit[$mc]=true;

		}

	}

}

function load_access_files($dir, &$access, $rec=0) {

	global $docroot, $conf, $access_cache, $os;

	if ($os=="win32") $dir=str_replace("\\", "/", $dir);
	
	if ($z=$access_cache[$dir]) {

		// Access cache hit
		
		$access=$z;
		return;
	
	} else {
	
		// Access cache miss
		
		if (!$rec) $access=array();
		
		$ed=explode("/", $dir);
		array_pop($ed);
		$ndir=implode("/", $ed);

		if ($os=="win32") $ndir=str_replace("/", "\\", $ndir);

		if (strpos($ndir."/", $docroot)===0) load_access_files($ndir, $access, $rec+1);

		$afn=$conf["global"]["accessfile"][0];

		if ($accesstmp=@file($dir."/".$afn)) if ($tmp_access=parseconfig($accesstmp, true)) foreach ($tmp_access as $key=>$val_arr) {
				
			foreach ($val_arr as $ckey=>$cval_arr) $access[$key][$ckey]=$cval_arr;
			if (!$rec) break;
		
		}

		$access_cache[$dir]=$access;
	
	}

}

function access_query($key, $idx=false) {

	global $access, $conf, $vhost;

	is_array($tmp=$access["global"][$key]) or
	is_array($tmp=$conf[$vhost][$key]) or
	$tmp=$conf["global"][$key];

	if ($idx===false) {

		return($tmp);

	} else {

		return($tmp[$idx]);
	
	}

}

function core_modules_hook($hname) {

	global $modules;

	if ($mh_arr=&$modules["core_".$hname]) {

		for($a=0;$a<count($mh_arr);$a++) {

			$mh_arr[$a]->main();
		
		}
	
	}

}

function log_srv($str) {

	global $conf;
	
	if ($s=$conf["global"]["serverlog"][0]) {

		if ($sl=@fopen($s, NW_BSAFE_APP_OPEN)) {

			fputs($sl, $str);
			fclose($sl);
			
		}

	}

}

function nw_gethostbyaddr($ip) {

	static $ns_lastreq;
	if ($ip!=$ns_lastreq[0]) $ns_lastreq=array($ip, @gethostbyaddr($ip));
	return($ns_lastreq[1]);

}

function _genpage_signature() {

	return(str_replace("%SERVERNAME", $GLOBALS["conf"][$GLOBALS["vhost"]]["servername"][0], GENPAGE_SIGNATURE));
	
}

function nw_server_vars() {

	global $conf;
	
	$filename=$GLOBALS["docroot"].$GLOBALS["http_uri"];
	
	$nsv["SERVER_SOFTWARE"]=SERVER_STRING;
	$nsv["SERVER_NAME"]=$conf[$GLOBALS["vhost"]]["servername"][0];
	$nsv["SERVER_PROTOCOL"]=HTTP_VERSION;
	$nsv["SERVER_PORT"]=$conf["global"]["listenport"][0];
	$nsv["SERVER_ADDR"]=$conf["global"]["listeninterface"][0];
	$nsv["SERVER_API"]=VERSION;
	$nsv["SERVER_ADMIN"]=$conf[$GLOBALS["vhost"]]["serveradmin"][0];
	$nsv["REQUEST_METHOD"]=$GLOBALS["http_action"];
	$nsv["PATH_TRANSLATED"]=$nsv["SCRIPT_FILENAME"]=realpath($filename);
	$nsv["SCRIPT_NAME"]="/".$GLOBALS["http_uri"];
	$nsv["QUERY_STRING"]=$GLOBALS["query_string"];
	$nsv["REMOTE_HOST"]=$GLOBALS["remote_host"];
	$nsv["REMOTE_ADDR"]=$GLOBALS["remote_ip"];
	$nsv["REMOTE_PORT"]=$GLOBALS["remote_port"];
	$nsv["AUTH_TYPE"]=$GLOBALS["auth_type"];
	$nsv["DOCUMENT_ROOT"]=$GLOBALS["docroot"];
	$nsv["REQUEST_URI"]="/".$GLOBALS["real_uri"].($nsv["QUERY_STRING"]?("?".$nsv["QUERY_STRING"]):"");
	$nsv["PATH_INFO"]=$GLOBALS["path_info"];

	foreach (access_query("addservervar") as $str) {

		$k=substr($str, 0, strpos($str, " "));
		$v=substr($str, strpos($str, " ")+1);
		if ($k) $nsv[$k]=$v;
	
	}

	return($nsv);

}

function techo($s, $flush=false) {

	global $conf;
	static $srv_buf;
	
	$tl=date("Ymd-His")." - $s\n";

	if (!$conf && !$flush) {

		$srv_buf.=$tl;
	
	} else {
	
		if ($srv_buf) {

			$tl=$srv_buf.$tl;
			$srv_buf="";

		}

		if ($conf["global"]["servermode"][0]!="inetd") {
		
			echo $tl;
			flush();

		}

		log_srv($tl);

	}

}

function errexit($s) {

	global $pidfile;
	
	techo("FATAL: ".$s, true);
	if ($pidfile) unlink($pidfile);
	exit;

}

function url_to_absolute($url) {

	$lport=$GLOBALS["conf"]["global"]["listenport"][0];
	return("http://".$GLOBALS["conf"][$GLOBALS["vhost"]]["servername"][0].(($lport!=80)?(":".$lport):"").($url!="/"?"/":"").$url);

}

function loadfile($filename, $extension, &$rq_err, &$cgi_headers, &$add_ctype, $force_parser=false) {
	
	global $conf, $docroot, $modules, $htreq_headers, $rq_file;
	
	$add_ctype=false;

	if (is_link($filename)) $filename=readlink($filename);
	chdir(dirname(realpath($filename)));
	$filename=basename($filename);

	if (($parser=trim($conf["global"]["_parseext"][strtolower($extension)])) || ($parser=$force_parser)) {

		// Parsed content
		
		if (strpos($parser, " ")!==false) {
		
			$ps_type=substr($parser, 0, strpos($parser, " "));
			$ps_arg=substr($parser, strpos($parser, " ")+1);
			if (strpos($ps_arg, '$')!==false) foreach (nw_server_vars() as $nkey=>$nval) $ps_arg=str_replace('$'.$nkey, $nval, $ps_arg);

		} else $ps_type=$parser;

	} else {

		// Static content

		$ps_type=$conf["global"]["defaulthandler"][0];

	}

	if ($ps=&$modules["parser_".$ps_type][0]) {

		$ps->parser_open($ps_arg, $filename, $rq_err, $cgi_headers, $add_ctype);

	} else {

		$rq_err=500;
		$GLOBALS["add_errmsg"]="Unable to find an appropriate parser for this content type.<br><br>";
		$ps=$GLOBALS["null_response"];
	
	}

	return($ps);

}

function nw_use_chunked_encoding() {

	global $lf, $keepalive, $http_version;
	
	if (!isset($lf->content_length) && $keepalive) {

		if ($http_version>="1.1") {

			return(true);
		
		} else {

			return("CLOSE");
		
		}
	
	} else return(false);

}

function build_response_headers() {

	global $http_resp, $HTTP_HEADERS, $rq_err, $add_ctype, $out_contenttype, $out_add_headers, $keepalive, $conf, $chunked, $lf;
	
	$out_headers=HTTP_VERSION." ".($http_resp?trim($http_resp):$HTTP_HEADERS[$rq_err])."\r\n";
	$out_headers.="Date: ".gmdate("D, d M Y H:i:s T")."\r\n";
	$out_headers.="Server: ".SERVER_STRING." (PHP/".phpversion().")\r\n";
	if (!$add_ctype) $out_headers.="Content-Type: ".$out_contenttype."\r\n";
	unset($add_ctype);

	if ($ahlist=access_query("addheader")) foreach ($ahlist as $val) $out_headers.=trim($val)."\r\n";
	if (($rq_err>=400) && ($eh=access_query("_errorheader", $rq_err))) $out_headers.=$eh."\r\n";
	
	if ($out_add_headers) {
		
		foreach ($out_add_headers as $key=>$val) if ($key=="cookies") {
			
			foreach ($val as $cval) $out_headers.="Set-Cookie: ".$cval."\r\n";
		
		} else {
			
			$out_headers.=$key.": ".$val."\r\n";

		}

		unset($out_add_headers);
	
	}

	if ($keepalive) {
	
		$out_headers.="Connection: Keep-Alive\r\n";
		$out_headers.="Keep-Alive: timeout=".(int)$conf["global"]["requesttimeout"][0].", max=".(int)($conf["global"]["keepalive"][0])."\r\n";
		
	} else {
		
		$out_headers.="Connection: close\r\n";

	}
	
	if ($chunked) { 
		
		$out_headers.="Transfer-Encoding: chunked\r\n";

	} else {

		if (is_int($lf->content_length)) $out_headers.="Content-Length: ".$lf->content_length."\r\n";

	}

	return($out_headers);

}

function nanoweb_init($conffile) {

	global $conf, $cmdline_conf_overrides, $cmdline_conf_adds, $modules, $posix_av, $pcntl_av, $gz_av, $mime;
	
	$conf=parseconfig(file($conffile));
	$conf=cmdline_conf_upd($conf, $cmdline_conf_overrides, $cmdline_conf_adds);
	$modules=load_modules($conf);
	modules_init();

	$posix_av=is_callable("posix_setuid");
	$pcntl_av=is_callable("pcntl_fork");
	$gz_av=is_callable("gzencode");

	if ((!$conf["global"]["singleprocessmode"][0]) && (!$posix_av || !$pcntl_av || ($conf["global"]["servermode"][0]=="inetd"))) {

		techo("WARN: forcing single process mode");
		$conf["global"]["singleprocessmode"][0]=true;

	}

	if ($conf["global"]["servermode"][0]=="inetd") {
		
		unset($conf["global"]["logtoconsole"]);
		unset($conf["global"]["pidfile"]);

	}

	if ($conf["global"]["singleprocessmode"][0]) {
		
		$conf["global"]["loggerprocess"]=0;
		$conf["global"]["keepalive"]=0;

	}

	if ($pcntl_av) {

		pcntl_signal(SIGTERM, "nanoweb_shutdown");
		pcntl_signal(SIGHUP, "nanoweb_reload");

	}

	$mime=array();
	
	if (!@is_readable($conf["global"]["mimetypes"][0])) {

		techo("WARN: unable to read mime types file (".$conf["global"]["mimetypes"][0]."), using internals");

		$mime["html"]="text/html";
		$mime["gif"]="image/gif";
		$mime["jpeg"]="image/jpeg";
		$mime["png"]="image/png";
		$mime["tgz"]="application/gtar";

	} else if ($mimetypes=@file($conf["global"]["mimetypes"][0])) {
		
		while (list($key, $s)=each($mimetypes)) if (trim($s) && ($s[0]!="#")) if (ereg("([a-zA-Z0-9/.-]+)[ \t]+([a-zA-Z0-9 -]+)", $s, $res)) if ($exts=explode(" ", trim($res[2]))) while (list($key2, $ext)=each($exts)) if (trim($res[1]) && trim($ext)) $mime[$ext]=trim($res[1]);

		unset($mimetypes);

	}

}

function nanoweb_shutdown($sig_no=SIGTERM) {

	global $sock, $mypid, $pidfile, $loggers_sck, $conf, $pmode;
	
	if ($pmode=="master") {
	
		if ($sock) {
			
			socket_shutdown($sock);
			socket_close($sock);

		}

		if ($nb_loggers=$conf["global"]["loggerprocess"][0]) {
		
			techo("halting loggers");
			
			for ($a=0;$a<$nb_loggers;$a++) {

				$pkt=NM_SEP_MAGIC."TERM";
				
				socket_write($loggers_sck, $pkt);
				usleep(100000);
			
			}

			sleep(1);

		}

		if (!$mypid) techo("daemon stopped\n");
		if ($pidfile) unlink($pidfile);

	}

	exit;

}

function nanoweb_reload($sig_no=SIGHUP) {

	global $mypid, $conffile;

	if (!$mypid) {
		
		nanoweb_init($conffile);
		techo("configuration file (".$conffile.") reloaded"); 

	}

}

function read_request(&$sck_connected, &$dp, &$pn, $maxlen=0) {
			
	global $pfdset, $conf, $msgsock, $inetd_in;
	static $rr_buffer;
	
	$wstart=(int)date("U");
		
	if ($rr_buffer!=="") {
		
		$buf=$rr_buffer;
		$tnreq=true;

	}
	
	while (!$rq_finished && $sck_connected) {
	
		if (!$tnreq) {
		
			$fdset=$pfdset;

			if ($conf["global"]["servermode"][0]=="inetd") {

				// Inetd

				if (feof($inetd_in)) {
					
					$tmp=false;
					$sck_connected=false;

				} else $tmp=fgetc($inetd_in);

			} else {
			
				if ($ns=socket_select($fdset, $write=NULL, $fdset, SCK_READ_SELECT_TIMEOUT)) $tmp=@socket_read($msgsock, SCK_READ_PACKET_SIZE); else $tmp=false;

			} 

		}

		if ($tmp || $tmp==="0" || $tnreq) {
			
			$tnreq=false;
			$wstart=(int)date("U");
			$buf.=$tmp;
			$pn=0;

			if (!$maxlen) {
			
				if (!$rnloop) {
					
					$buf=ltrim($buf);
					$rnloop=true;

				}

				if (($dp=strpos($buf, "\r\n\r\n"))!==false) $pn=4;
				else if (($dp=strpos($buf, "\n\n"))!==false) $pn=2;

				if ($pn) $rq_finished=true;

			} else {

				if (strlen($buf)>=$maxlen) $rq_finished=true;

			}

		} else if (($ns) || ((!$ns) && (((int)date("U")-$wstart)>=$conf["global"]["requesttimeout"][0]))) $sck_connected=false;

	}
	
	if (!$maxlen) {

		$tbuf=substr($buf, 0, $dp+$pn);
		$rr_buffer=substr($buf, $dp+$pn);
		$buf=$tbuf;
	
	} else {

		$tbuf=substr($buf, 0, $maxlen);
		$rr_buffer=substr($buf, $maxlen);
		$buf=$tbuf;

	}
	
	return($buf);

}

function send_response($response, &$sck_connected) {

	global $msgsock, $pfdset, $conf;
	
	$resp_len=strlen($response);

	while($sent_len<$resp_len && $sck_connected) {

		if ($conf["global"]["servermode"][0]=="inetd") {

			// Inetd
			
			echo $response;
			$sent_len=strlen($response);

		} else {

			$fdset=$pfdset;
			
			if (($sent_len+SCK_WRITE_PACKET_SIZE)>$resp_len) $size=$resp_len-$sent_len; else $size=SCK_WRITE_PACKET_SIZE;
			if (($ret=@socket_write($msgsock, substr($response, $sent_len, $size), $size))>0) $sent_len+=$ret;
			if (!socket_select($read=NULL, $fdset, $except=NULL, 10)) $sck_connected=false;

		}

	}

	return($sent_len);

}

function int_sendtomaster($msg) {

	global $master_sck, $conf;

	if (!$conf["global"]["singleprocessmode"][0]) {
	
		$pkt=NM_SEP_MAGIC.$msg;
		
		$ret=socket_write($master_sck, $pkt);

		if ($ret!=strlen($pkt)) {

			techo("WARN: master process is not responding");
		
		}

	}

}

function _server_report_state($s, $remote_host="") {

	$tmp[0]=$GLOBALS["mypid"];
	$tmp[1]=$s;

	if ($remote_host) $tmp[2]=$remote_host;

	int_sendtomaster(NM_SERVER_STATE.serialize($tmp));

}

function logger_run($logger_id) {

	global $conf, $children_logsck, $modules, $plgset, $pmode;

	$pmode="logger";

	pcntl_signal(SIGTERM, SIG_DFL);
	pcntl_signal(SIGHUP, SIG_IGN);

	$mypid=posix_getpid();

	if ($setgid=@posix_getgrnam($conf["global"]["loggergroup"][0])) posix_setgid($setgid["gid"]);
	else if ($setgid=@posix_getgrnam($conf["global"]["group"][0])) posix_setgid($setgid["gid"]);

	if ($setuid=@posix_getpwnam($conf["global"]["loggeruser"][0])) posix_setuid($setuid["uid"]);
	else if ($setuid=@posix_getpwnam($conf["global"]["user"][0])) posix_setuid($setuid["uid"]);
	
	techo("logger process #".$logger_id." is running (pid=".$mypid.")");

	while (!$logger_exit) {
		
		$r=socket_read($children_logsck, INT_MSGSIZE);

		$msg=substr($r, NM_SEP_LEN);

		if (strpos($msg, NM_SEP_MAGIC)) {

			$msgs=explode(NM_SEP_MAGIC, $msg);
		
		} else $msgs=array($msg);

		foreach($msgs as $r) switch($r) {

			case "TERM": 
			$logger_exit=true;
			break;

			default:
			$l=unserialize($r);

			// Reverse DNS query if the server hasn't done it before
			
			if (($conf["global"]["hostnamelookups"][0]) && ($conf["global"]["hostnamelookupsby"][0]=="logger") && ($rhost=nw_gethostbyaddr($l[2]))) {
				
				$l[1]=$rhost;

			} else {
			
				$l[1]=$l[2];

			}
			
			if ($nb_log_mods=count($modules["log"])) for ($a=0;$a<$nb_log_mods;$a++) $modules["log"][$a]->log_hit($l[0], $l[1], $l[2], $l[3], $l[4], $l[5], $l[6], $l[7], $l[8]);
			break;
			
		}
	
	}

	techo("logger process #".$logger_id." stopped");
	exit;

}

function spawn_loggers($nb_loggers) {
	
	static $logger_id;
	global $logger_pids;
	
	for ($a=0;$a<$nb_loggers;++$a) {

		$pid=pcntl_fork();
		$logger_id++;		

		if ($pid===0) {
			
			logger_run($logger_id);

		} else {
			
			$logger_pids[$pid]=$logger_id;
			if ($nb_loggers>1) usleep(100000);

		}
			
	}

}

// Begin

set_time_limit(0);

techo("aEGiS nanoweb/".VERSION." (c)2002 by sIX <six@aegis-corp.org>");

$stats_start=date("U");

if (version_compare(phpversion(), REQUIRED_PHP_VERSION)<0) errexit("nanoweb needs PHP >= ".REQUIRED_PHP_VERSION." (you are using ".phpversion().")");
if (version_compare(phpversion(), "4.3.0")>=0) $sckv3=true;

$os=strpos(strtolower(PHP_OS), "win")===false?"unix":"win32";

switch ($os) {

	case "win32":
	define(NW_BSAFE_READ_OPEN, "rb");
	define(NW_BSAFE_WRITE_OPEN, "wb");
	define(NW_BSAFE_APP_OPEN, "ab");
	break;

	default:
	define(NW_BSAFE_READ_OPEN, "r");
	define(NW_BSAFE_WRITE_OPEN, "w");
	define(NW_BSAFE_APP_OPEN, "a");
	break;

}

while (list($f_name, $f_mandatory)=each($TEST_FUNCS)) if (!is_callable($f_name)) {

	if ($f_mandatory) errexit("function '".$f_name."' not available, aborting");
	else techo("WARN: function '".$f_name."' not available");

}

// Parse command line

$cmdline_help[]="usage: ./nanoweb.php [/path/to/nanoweb.conf] [options]";
$cmdline_help[]="";
$cmdline_help[]="nanoweb supported command line options :";
$cmdline_help[]="";
$cmdline_help[]="--help : this help screen";
$cmdline_help[]="--version : show version info";
$cmdline_help[]="--config=/path/to/nanoweb.conf : configuration file";
$cmdline_help[]="--set-option=\"optionname=optionvalue\" : set configuration option";
$cmdline_help[]="--add-option=\"optionname=optionvalue\" : add configuration option";
$cmdline_help[]="";
$cmdline_help[]="can be abreviated as -h, -v, -c, -o and -a";
$cmdline_help[]="";

if ($_SERVER["argc"]>1) for($a=1;$a<$_SERVER["argc"];$a++) {

	if (($a==1) && (substr($_SERVER["argv"][$a], 0, 1)!="-")) {

		$cmdline_cf=$_SERVER["argv"][1];
	
	} else {

		$ca=explode("=", $_SERVER["argv"][$a]);
		$ck=array_shift($ca);
		$cv=implode("=", $ca);

		switch($ck) {

			case "-?":
			case "-h":
			case "--help":
			foreach ($cmdline_help as $cs) echo($cs."\n");
			exit;
			break;

			case "-v":
			case "--version":
			die(VERSION."\n");
			break;
			
			case "-c":
			case "--config":
			$cmdline_cf=$cv;
			break;

			case "-o":
			case "--set-option":
			$cmdline_conf_overrides[]=$cv;
			break;

			case "-a":
			case "--add-option":
			$cmdline_conf_adds[]=$cv;
			break;

			default:
			errexit("unknown argument : ".$_SERVER["argv"][$a].", try --help");
			break;

		}
	
	}
	
}

unset($cmdline_help);

if ($cmdline_cf) $conffile=$cmdline_cf; else $conffile=DEFAULT_CONFIG_FILE;

if (!is_readable($conffile)) {
	
	foreach ($cmdline_help as $s) techo($s);
	errexit("unable to read configuration (".$conffile."), aborting");

}

nanoweb_init($conffile);

if ($posix_av && $conf["global"]["pidfile"][0]) {
	
	$pidfile=$conf["global"]["pidfile"][0];
	
	if ($f=fopen($pidfile, "w")) {
	
		fputs($f, (string)posix_getpid()."\n");
		fclose($f);

	} else {

		techo("WARN: unable to open pid file '".$pidfile."'");
		unset($pidfile);
	
	}

}

if ($conf["global"]["servermode"][0]!="inetd") {

	if (($sock = @socket_create(AF_INET, SOCK_STREAM, 0))<0) errexit("socket_create() failed: ".socket_strerror(socket_last_error()));

	if ($sckv3) socket_set_option($sock, SOL_SOCKET, SO_REUSEADDR, 1);
	else socket_setopt($sock, SOL_SOCKET, SO_REUSEADDR, 1);

	if (!@socket_bind($sock, $conf["global"]["listeninterface"][0], $conf["global"]["listenport"][0])) errexit("socket_bind() failed: ".socket_strerror(socket_last_error()));
	if (!@socket_listen($sock, $conf["global"]["listenqueue"][0])) errexit("socket_listen() failed: ".socket_strerror(socket_last_error()));

	socket_set_nonblock($sock);

}

$pmode="master";

if ($nb_loggers=$conf["global"]["loggerprocess"][0]) {

	// Prepare and spawn logger processes if specified
	
	techo("spawning loggers");

	$sck_pair=array();
	socket_create_pair(AF_UNIX, SOCK_STREAM, 0, $sck_pair);

	$children_logsck=&$sck_pair[0];
	$loggers_sck=&$sck_pair[1];
	
	spawn_loggers($nb_loggers);

} else {

	// Be sure not to ask anything to loggers

	$conf["global"]["hostnamelookupsby"][0]="server";

}

if ($pcntl_av) {

	$sck_pair=array();
	socket_create_pair(AF_UNIX, SOCK_STREAM, 0, $sck_pair);

	$children_sck=&$sck_pair[0];
	$master_sck=&$sck_pair[1];

	socket_set_nonblock($children_sck);
	socket_set_nonblock($master_sck);

}

$plnset=array($sock, $children_sck);

if ($conf["global"]["servermode"][0]!="inetd") {

	techo("running and accepting connections on port ".$conf["global"]["listenport"][0]);

} else {

	$inetd_in=fopen("php://stdin", NW_BSAFE_READ_OPEN);
	set_file_buffer($inetd_in, 0);

	techo("running in inetd mode");

}

$def_cnx=($conf["global"]["servermode"][0]=="inetd");

while (true) {

	$cnx=$def_cnx;
	
	while (!$cnx) {

		declare (ticks = 1) {
		
			// Allow to catch signals here			
			
			$lnset=$plnset;

		}
		
		$ns=socket_select($lnset, $write=NULL, $except=NULL, 1);
		
		if ($ns) {
				
			switch ($lnset[0]) {
			
				case $sock:

				if (is_resource($msgsock=@socket_accept($sock))) {
				
					// We have a new connection
					
					$stats_cnx++;
					$cnx=true;

				}
				
				break;


				case $children_sck:
				
				if ($msg=socket_read($children_sck, INT_MSGSIZE)) {

					// Message from a child process
					
					$msg=substr($msg, NM_SEP_LEN);

					if (strpos($msg, NM_SEP_MAGIC)) {

						$msgs=explode(NM_SEP_MAGIC, $msg);
					
					} else $msgs=array($msg);

					foreach($msgs as $msg) {

						$mtype=substr($msg, 0, 5);
						$mcontent=substr($msg, 5);
						
						switch ($mtype) {

							case NM_HIT: 
								
							$mtmp=unserialize($mcontent);
							$stats_resperr[$mtmp["status"]]++;
							$stats_total_size+=$mtpm["length"];
							$stats_vhosts[$conf[$mtmp["vhost"]]["servername"][0]]++;
							$stats_xfer+=(float)$mtmp["length"];
							$stats_hits++;
							
							break;


							case NM_RESTART_LOGGERS:

							techo("respawning loggers");
							spawn_loggers($conf["global"]["loggerprocess"][0]);
							
							break;


							case NM_SERVER_STATE:

							$mtmp=unserialize($mcontent);
							
							if ($server_pids[$mtmp[0]]) {
							
								$server_pids[$mtmp[0]]=$mtmp[1];
								if ($mtmp[2]) $server_pids_remote[$mtmp[0]]=$mtmp[2];

							}
							
							break;
						
						}

					}

				}

				
				break;
			
			}
		
		} 

		if (!$conf["global"]["singleprocessmode"][0]) if ($deadpid=pcntl_waitpid(-1, $cstat, WNOHANG)) {

			// Dead child

			if (($dead_logger=$logger_pids[$deadpid]) && ($conf["global"]["loggerprocess"][0])) {

				// Dead logger (this is abnormal, we have to restart one)

				techo("logger process #".$dead_logger." died (pid=".$deadpid."), respawning");
				unset($logger_pids[$deadpid]);
				spawn_loggers(1);
			
			} else {

				// Dead child server, clear servers tables

				unset($server_pids[$deadpid]);
				unset($server_pids_remote[$deadpid]);
			
			}
		
		}

	}
	
	if ($conf["global"]["singleprocessmode"][0]) {
		
		$pid=0;

		// Invalidate access cache every 5 connections
		if (($stats_cnx%5)==0) unset($access_cache);

	} else {
		
		$pid=pcntl_fork();

	}

	if ($pid===0) {
	
		if ($pcntl_av) {

			pcntl_signal(SIGTERM, SIG_DFL);
			pcntl_signal(SIGHUP, SIG_DFL);

		}
		
		if ($posix_av) $mypid=posix_getpid();
		
		set_time_limit($conf["global"]["childlifetime"][0]);
		
		if ($conf["global"]["servermode"][0]!="inetd") {
		
			socket_getpeername($msgsock, $remote_ip, $remote_port);
			socket_set_nonblock($msgsock);
			$pfdset=array($msgsock);

			if (($conf["global"]["hostnamelookups"][0]) && ($conf["global"]["hostnamelookupsby"][0]!="logger") && ($rhost=nw_gethostbyaddr($remote_ip))) {
				
				$remote_host=$rhost;

			} else {
			
				$remote_host=$remote_ip;

			}

		} else {

			$remote_ip=getenv("INETD_REMOTE_IP");
			$remote_port=getenv("INETD_REMOTE_PORT");
		
		}

		core_modules_hook("after_fork_server");
		
		while ($cnx) {
		
			_server_report_state("(waiting for request)", $remote_host);
			
			$sck_connected=true;

			$buf=read_request($sck_connected, $dp, $pn);
			
			unset($pri_redir);

			if ($sck_connected) {

				if (strlen($buf)!=$dp+4) $add_req=substr($buf, $dp+4); else $add_req="";
				$tmp_arr=explode("\n",substr($buf, 0, $dp));

				$htreq=explode(" ", trim($tmp_arr[0]));
				$http_action=strtoupper($htreq[0]);
				$http_protocol=explode("/", $htreq[2]);
				$http_version=$http_protocol[1];

				// Decode HTTP request headers
				
				unset($htreq_headers);
				
				$tmparcnt=count($tmp_arr);
				for($a=1;$a<=$tmparcnt;$a++) if ($s=trim($tmp_arr[$a])) {

					$hd_key=trim(strtoupper(substr($s, 0, strpos($s, ":"))));
					$hd_val=trim(substr($s,strpos($s, ":")+1));

					$htreq_headers[$hd_key]=$hd_val;

				}

				// Decode Host header
					
				$host=trim(strtolower($htreq_headers["HOST"]));
				if (($p=strpos($host, ":"))!==false) $host=substr($host, 0, $p);

				do {
				
					// Try vhost=host
					
					if (is_array($conf[$host])) {
							
						$vhost=$host;
						break;
						
					} 

					// Try wildcards

					$hlen=strlen($host);
					for($vhlen=0;$vhlen<=$hlen;$vhlen++) {
						
						$whost="*".substr($host, $vhlen);

						if (is_array($conf[$whost])) {

							$vhost=$whost;
							break 2;
						
						}

					}

					// Or set to global

					$vhost="global";

				} while (0);
				
				// Decode HTTP Authentication header

				if ($auth_hdr=$htreq_headers["AUTHORIZATION"]) {

					$dtmp=explode(" ", $auth_hdr);
					$auth_type=$dtmp[0];
					$auth_lp=explode(":", base64_decode($dtmp[1]));
					$auth_user=$auth_lp[0];
					$auth_pass=$auth_lp[1];
				
				} else $auth_type=$auth_user=$auth_pass="";
				
				// Decode Keep-Alive header
				
				$keepalive=(strtolower(trim($htreq_headers["CONNECTION"]))=="keep-alive" && (int)$conf["global"]["keepalive"][0]>1);
				if ($keepalive && ++$rq_count>=(int)$conf["global"]["keepalive"][0]) $keepalive=false;
				$cnx=$keepalive;
				
				// Set Uid and Gid

				if (!$uid_set) {
				
					$cfgid=$conf[$vhost]["group"][0];
					$cfuid=$conf[$vhost]["user"][0];
					
					if ($posix_av) {
					
						if ($cfgid) if ($setgid=@posix_getgrnam($conf[$vhost]["group"][0])) posix_setgid($setgid["gid"]);
						if ($cfuid) if ($setuid=@posix_getpwnam($conf[$vhost]["user"][0])) posix_setuid($setuid["uid"]);

					}

					$uid_set=true;
				
				}
				
				$complete_path=$http_uri=$rq_file=$out_headers="";
				$rq_err=0;
				
				$docroot=$conf[$vhost]["documentroot"][0];
				if ($docroot[strlen($docroot)-1]!="/") $docroot.="/";
				
				$http_protocol=$htreq[count($htreq)-1];
				
				if ($p1=$htreq[1]) {

					$p1=explode("?", $p1);
					$complete_path=$real_uri=ltrim($http_uri=urldecode($p1[0]), "/");
					$query_string=$p1[1];

					$hu=$docroot.$real_uri;

					// Load access files if needed
					
					unset($access);
					
					if (is_dir($hu)) {
						
						$uridir=$http_uri;
						
					} else if (is_dir($docroot.($uridn=dirname($http_uri)))) {

						$uridir=$uridn;
					
					} else $uridir="";

					if (($accessdir=realpath($docroot.$uridir)) && ($conf["global"]["accessfile"][0])) load_access_files($accessdir, $access);

					core_modules_hook("before_decode");
					
					foreach (access_query("_aliases") as $key=>$val) if (strpos($http_uri, $key)===0) {

						// Alias
						
						$docroot=$val.((substr($val, strlen($val)-1, 1)=="/")?"":"/");
						$http_uri=str_replace($key, "", $http_uri);

						if ((is_dir($docroot.$http_uri)) && (substr($docroot.$http_uri, strlen($docroot.$http_uri)-1, 1)!="/")) $pri_redir=$htreq[1]."/";

						break;

					}

					$http_uri=ltrim($http_uri, "/");
					
					if (($http_uri[0]=="~") && ($udadd=$conf[$vhost]["userdir"][0]) && ($posix_av)) {

						// User directory
						
						if (strpos($http_uri, "/")===false) {
							
							$pri_redir=$htreq[1]."/";

						} else {

							$udname=substr($http_uri, 1, strpos($http_uri, "/")-1);
							$udres=substr($http_uri, strpos($http_uri, "/")+1);
							
							if ($udinf=@posix_getpwnam($udname)) {

								$docroot=$udinf["dir"]."/".$udadd."/";
								$http_uri=$udres;
							
							}

						}
					
					}
					
					if (is_dir($docroot.$http_uri) && !$pri_redir) {
						
						if ($http_uri && substr($http_uri, -1)!="/") {

							$pri_redir=$htreq[1]."/";
						
						} else if ($dilist=access_query("directoryindex", 0)) {
						
							$dis=explode(" ", $dilist);

							foreach ($dis as $diname) if (is_readable($docroot.$http_uri.$diname)) {

								$http_uri.=$diname;
								break;

							}

						}

					}

					$path_info="";
				
					if (access_query("allowpathinfo", 0) && !file_exists($docroot.$http_uri)) {
						
						// Try path_info

						$new_uri=$http_uri;
						
						while (!@is_file($docroot.$new_uri) && $new_uri) {

							$new_uri=substr($new_uri, 0, strrpos($new_uri, "/"));

							if (!@is_file($docroot.$new_uri) && $pie_arr=access_query("pathinfotryext")) foreach ($pie_arr as $pie_ext) if (@is_file($docroot.$new_uri.".".$pie_ext)) {

								$new_uri.=".".$pie_ext;
								break;
							
							}

						}

						if ($new_uri) {

							// Path_info found

							$path_info=substr($http_uri, strlen($new_uri));
							$http_uri=$new_uri;
						
						}
					
					}
					
					$rq_file=pathinfo($http_uri);

				}

				$hbn=basename($http_uri);
				unset($bad_rq);
				
				if (realpath($docroot.$http_uri) && (strpos(realpath($docroot.$http_uri), realpath($docroot))===false)) {
					
					$bad_rq=NW_BAD_OUTSIDE_DOCROOT;

				} else if (($conf[$vhost]["ignoredotfiles"][0]) && ((substr($hbn, 0, 1)==".") && ($hbn!=".."))) {

					$bad_rq=NW_BAD_DOT_FILE;
				
				}

				if (($bad_rq==NW_BAD_OUTSIDE_DOCROOT) && ($conf["global"]["allowextsymlinks"][0]) && (is_link($docroot.$http_uri)) && (strpos(realpath(dirname($docroot.$http_uri)), realpath($docroot))===0)) unset($bad_rq);
				
				switch ($bad_rq) {

					case NW_BAD_OUTSIDE_DOCROOT:
					techo("NOTICE: discarded request outside of document root (".$docroot.$http_uri.")");
					$http_uri="";
					$pri_err=404;
					break;

					case NW_BAD_DOT_FILE:
					techo("NOTICE: discarded request for dot file (".$docroot.$http_uri.")");
					$http_uri="";
					$pri_err=404;
					break;

					default:

				}

				$sst=$http_action." http://".$htreq_headers["HOST"]."/".$real_uri.($query_string?("?".$query_string):"");
				_server_report_state($sst);
				
				if ($hu!=$docroot.$http_uri) {
				
					// Reload access files if needed
					
					$hu=$docroot.$http_uri;

					unset($access);
					
					if (is_dir($hu)) {
						
						$uridir=$http_uri;
						
					} else if (is_dir($docroot.($uridn=dirname($http_uri)))) {

						$uridir=$uridn;
					
					} else $uridir="";

					if (($accessdir=realpath($docroot.$uridir)) && ($conf["global"]["accessfile"][0])) load_access_files($accessdir, $access);

				}
				
				$out_contenttype=$default_ct=access_query("defaultcontenttype", 0);
				unset($out_add_headers);
				
				if ($htreq_headers["CONTENT-LENGTH"]) {

					// Read request content if there is one (POST requests)
					
					$buf=$add_req;
					if (strlen($buf)<$htreq_headers["CONTENT-LENGTH"]) $buf.=read_request($sck_connected, $dp, $pn, $htreq_headers["CONTENT-LENGTH"]-strlen($buf));
					$htreq_content=substr($buf, 0, $htreq_headers["CONTENT-LENGTH"]);
				
				}

				// Auth handler
				
				$logged_user="";
				
				if ($rauths=access_query("authrequire")) {
					
					foreach ($rauths as $rauth) {

						if ($spos=strpos($rauth, " ")) {
						
							$authtype=strtolower(substr($rauth, 0, strpos($rauth, " ")));
							$authargs=trim(substr($rauth, strpos($rauth, " ")));

						} else {

							$authtype=strtolower($rauth);
							$authargs="";

						}
						
						$authmodn="auth_".strtolower($authtype);

						if (is_object($modules[$authmodn][0])) {
							
							if ($modules[$authmodn][0]->auth($auth_user, $auth_pass, $authargs)) {

								$logged_user=$auth_user;
								break;
							
							}

						} else {

							techo("WARN: authentication module not found for type '".$authtype."'");
					
						}

					}
					
					if ($logged_user==="") {
					
						$logged_user=" ";
						$pri_err=401;
						$out_add_headers["WWW-Authenticate"]="Basic realm=\"".access_query("authrealm", 0)."\"";
						if ($emsg=access_query("authmessage", 0)) $add_errmsg.=$emsg."<br><br>";

					}
				
				}
				
				core_modules_hook("after_decode");

				if ($sck_connected) {

					switch ($http_action) {

						case "POST":

						if ((!$conf["global"]["_parseext"][strtolower($rq_file["extension"])]) && (is_file($docroot.$http_uri))) $pri_err=405;
						
						case "GET":
						case "HEAD":

						if ($pri_err) {

							// Internal setting of http error

							$rq_err=$pri_err;
							unset($pri_err);
						
						} else if ($pri_redir) {

							// Internal redirection

							if ($rq_err<300) $rq_err=302;
							$out_add_headers["Location"]=$pri_redir;
							if (version_compare($http_version, "1.0")<=0) $out_add_headers["URI"]=$pri_redir;

							unset($pri_redir);
						
						} else if (is_object($modules["url:/".$http_uri])) {

							// Module URL Hook
							
							$lf =& new static_response($modules["url:/".$http_uri]->url($rq_err, $out_contenttype, $out_add_headers));
						
						} else if (is_dir($docroot.$http_uri)) {

							// Directory without index

							$rq_err=404;
							core_modules_hook("directory_handler");

						} else if (!is_file($docroot.$http_uri)) {
							
							// 404 Not Found

							$rq_err=404;

						} else if (!is_readable($docroot.$http_uri)) {
							
							// 403 Forbidden
							
							$rq_err=403;

						} else  {

							// 200 OK
							
							$rq_err=200;
							$lf=loadfile($docroot.$http_uri, $rq_file["extension"], $rq_err, $out_add_headers, $add_ctype, $pri_parser);

							/* libphpHACK
							 * 
							 * uncomment the code below to experiment with mod_libphp
							
							if (isset($__nw_libphp_script)) { 
								
								unset($lf);
								include($__nw_libphp_script); 
								exit; 
								
							}

							*/
							
							if ($mimetype=$mime[strtolower($rq_file["extension"])]) $out_contenttype=$mimetype; else $out_contenttype=$default_ct;
							unset($pri_parser);

						}

						if ($http_action=="HEAD") {
							
							$plen=$lf->content_length;
							$lf=$null_response;
							$lf->content_length=$plen;

						}

						break;

						case "OPTIONS":
						$rq_err=200;
						$out_add_headers["Allow"]="GET, ".($conf["global"]["_parseext"][strtolower($rq_file["extension"])]?"POST, ":"")."HEAD, OPTIONS";
						break;

						case "PUT":
						case "DELETE":
						case "TRACE": $rq_err=501;
						break;

						default: $rq_err=400;
						break;

					}

					if ($rq_err!=200 && $rq_err!=416) {

						// Error messages
						
						if ($rq_err>=400) {

							if (($errordoc=trim(access_query("_errordocument", $rq_err))) && (@is_readable($docroot.$errordoc))) {

									$http_uri=$errordoc;
									$errext=substr(strrchr($errordoc, "."), 1);
									$lf=loadfile($docroot.$errordoc, $errext, $rq_err, $out_add_headers, $add_ctype);

									if ($mimetype=$mime[strtolower($errext)]) $out_contenttype=$mimetype; else $out_contenttype=$default_ct;
								
							} else {
							
								$out_contenttype="text/html";
								$err_page="<html><head><title>".$HTTP_HEADERS[$rq_err]."</title></head><body><h1>".$HTTP_HEADERS[$rq_err]."</h1>";
								if ($real_uri) $err_page.="An error occured trying to access document or resource <b>/".htmlentities($real_uri)."</b>.<br><br>";
								
								if ($add_errmsg) {

									$err_page.=$add_errmsg;
									unset($add_errmsg);
								
								}

								if ($admmail=$conf[$vhost]["serveradmin"][0]) $err_page.="Please contact the server <a href=\"mailto:".$admmail."\">administrator</a> if the problem is abnormal and persists.<br><br>";
								$err_page.="<hr style=\"height:1px\"><font size=\"1\">"._genpage_signature()."</font></body></html>";

								$lf =& new static_response($err_page);

							}
						
							$cnx=false;
						
						} else if ($rq_err>=301) {

							$lf=$null_response;
						
						}
					
					}
					
					core_modules_hook("before_response");
					
					if (!$rq_err) $rq_err=500;
					
					$chunked=nw_use_chunked_encoding();
					if ($chunked==="CLOSE") $cnx=$keepalive=$chunked=false;

					// Send the response headers and content

					$sent_len=0;
					$first_chunk=true;

					while (($buf=$lf->parser_get_output()) || $first_chunk) {

						if ($first_chunk) {

							$hbuf=build_response_headers()."\r\n";
						
						}
						
						if ($chunked) {

							$chunk_header=dechex(strlen($buf))."\r\n";
							$metasize=strlen($chunk_header)+2;
							$rbytes=send_response($hbuf.($buf?$chunk_header.$buf."\r\n":""), $sck_connected);
							$sent_len+=($rbytes-$metasize);

						} else {

							$sent_len+=send_response($hbuf.$buf, $sck_connected);
						
						}

						if ($first_chunk) {
						
							$sent_len-=strlen($hbuf);
							$hbuf="";
							$first_chunk=false;

						}
						
						if ($lf->parser_eof() || !$sck_connected) break;

					}

					$lf->parser_close();

					if ($chunked) {
						
						$meta_len=0;
						send_response("0\r\n\r\n", $sck_connected);

					}
		
					if (!$sck_connected) $cnx=false;
					if (($sent_content_length=$sent_len-$meta_len)<0) $sent_content_length=0;

					$hreport=serialize(array("status" => $rq_err, "content-length" => $sent_content_length, "length" => $sent_len, "vhost" => $vhost));
					int_sendtomaster(NM_HIT.$hreport);
					
					if ($conf["global"]["loggerprocess"][0]) {
					
						// Send the logging infos to dedicated processes
						
						$log_arr=array($vhost, $remote_host, $remote_ip, $logged_user, trim($tmp_arr[0]), $rq_err, $sent_content_length, $htreq_headers["REFERER"], $htreq_headers["USER-AGENT"]);

						$logmsg=serialize($log_arr);

						if (strlen($logmsg)>INT_MSGSIZE) {

							techo("WARN: internal communication error (packet too long)");
						
						} else {

							$pkt=NM_SEP_MAGIC.$logmsg;
							
							$r=socket_write($loggers_sck, $pkt);

							if ($r!=strlen($pkt)) {

								techo("WARN: unable to communicate with logger process");
								int_sendtomaster(NM_RESTART_LOGGERS);
							
							}

						}

					} else {

						// Or do it ourselves
					
						if ($nb_loggers=count($modules["log"])) {

							for ($a=0;$a<$nb_loggers;$a++) $modules["log"][$a]->log_hit($vhost, $remote_host, $remote_ip, $logged_user, trim($tmp_arr[0]), $rq_err, $sent_content_length, $htreq_headers["REFERER"], $htreq_headers["USER-AGENT"]);

						}

					}

					$lf=$null_response;
				}
			
			} else {

				$cnx=false;
			
			}

		}
		
		socket_shutdown($msgsock);
		socket_close($msgsock);
		
		if ((!$conf["global"]["singleprocessmode"][0]) || ($conf["global"]["servermode"][0]=="inetd")) exit;
	
	} else if ($pid==-1) {

		// Fork failed
		
		techo("ERROR: unable to pcntl_fork()");

	} else {

		// Fork successful
		
		$server_pids[$pid]="(connection accepted)";
		$server_pids_remote[$pid]="(unknown)";

		core_modules_hook("after_fork_master");
	
	}
	
}

?>
